

#if	defined(HASNCACHE)
/*
 * rnam.frag - read BSD format (struct namecache or nch) name cache
 *
 * This fragment is effective only when HASNCACHE is defined.
 */

/*
 * The caller must:
 *
 *	#include the relevant header file -- e.g., <sys/namei.h>.
 *
 *	Define X_NCACHE as the Nl[] index to the kernel cache address.
 *
 *	Define X_NCSIZE as the Nl[] index to the size of the kernel
 *	cache.
 *
 *	Define NCACHE_NXT if the kernel's name cache is a linked
 *	list starting at Nl[X_NCACHE].n_value, rather than a table
 *	starting at that address.
 *
 *	Define the name of the name cache structure -- e.g.,
 *
 *		#define NCACHE	<structure name>
 *
 *	Define the names of these elements of struct NCACHE:
 *
 *	must		#define NCACHE_NM	<name>
 *	must		#define NCACHE_NMLEN	<name length
 *	optional	#define NCACHE_NXT	<link to next entry>
 *	must		#define NCACHE_NODEADDR	<node address>
 *	must		#define NCACHE_NODEID	<node capability ID)
 *	optional	#define NCACHE_PARADDR	<parent node address>
 *	optional	#define NCACHE_PARID	<parent node capability ID)
 *
 * The caller may need to:
 *
 *	Define NCHNAMLEN as the length of the name element of NCACHE, if it's
 *	not defined in the header file that defines the NCACHE structure.
 *
 *	Define this prototype for ncache_load() if it is called
 *	before rnam.frag is included.
 *
 *		_PROTOTYPE(static void ncache_load,(void));
 */


/*
 * Local static values
 */

static int Mch;				/* name cache hash mask */
static int Nc = 0;			/* size of name cache */
static int Nch = 0;			/* size of name cache hash pointer
					 * table */
struct l_nch {
	unsigned long na;		/* node address */
	unsigned long id;		/* capability ID */

# if	defined(NCACHE_PARADDR) && defined(NCACHE_PARID)
	unsigned long pa;		/* parent node address */
	unsigned long did;		/* parent capability ID */
# endif	/* defined(NCACHE_PARADDR) && defined(NCACHE_PARID) */

	char nm[NCHNAMLEN+1];		/* name */
	int nl;				/* name length */
};

static struct l_nch *Ncache = NULL;	/* the local name cache */
static struct l_nch **Nchash = NULL;	/* Ncache hash pointers */

#define ncachehash(i,n)		Nchash+(((((int)(n)>>2)+((int)(i)))*31415)&Mch)

_PROTOTYPE(static struct l_nch *ncache_addr,(unsigned long i, unsigned long n));


/*
 * ncache_addr() - look up a node's local ncache address
 */

static struct l_nch *
ncache_addr(i, n)
	unsigned long i;		/* node's capability ID */
	unsigned long n;		/* node's address */
{
	struct l_nch **hp;

	for (hp = ncachehash(i, n); *hp; hp++) {
		if ((*hp)->id == i && (*hp)->na == n)
			return(*hp);
	}
	return(NULL);
}


/*
 * ncache_load() - load the kernel's name cache
 */

static void
ncache_load()
{
	struct l_nch **hp, *lc;
	int i, len, n;
	struct NCACHE *kc;
	KA_T kp;

# if	defined(NCACHE_NXT)
	struct NCACHE nc;
# else	/* !defined NCACHE_NXT) */
	struct NCACHE *kca;
# endif	/* defined(NCACHE_NXT) */

	if (!Fncache)
		return;
/*
 * Read cache size and cache address.
 */
	if (!Nl[X_NCSIZE].n_value
	||  kread((KA_T)Nl[X_NCSIZE].n_value, (char *)&Nc, sizeof(Nc)))
	{
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: can't read name cache size: %#x\n",
		    Pn, Nl[X_NCSIZE].n_value);
	    Nc = 0;
	    return;
	}
	if (Nc < 1) {

no_cache:

	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: unusable name cache size: %d\n", Pn, Nc);
	    Nc = 0;
	    return;
	}
	if (!Nl[X_NCACHE].n_value
	||  kread((KA_T)Nl[X_NCACHE].n_value, (char *)&kp, sizeof(kp))) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: can't read name cache pointer: %#x\n",
		    Pn, Nl[X_NCACHE].n_value);
	    Nc = 0;
	    return;
	}
/*
 * Allocate space for cache copies.
 */

# if	!defined(NCACHE_NXT)
	len = Nc * sizeof(struct NCACHE);
	if ((kca = (struct NCACHE *)malloc((MALLOC_S)len)) == NULL) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: can't allocate name cache space: %d\n", Pn, len);
	    exit(1);
	}
	if (kread(kp, (char *)kca, len)) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: can't read name cache: %#x\n", Pn, kp);
	    exit(1);
	}
# endif	/* !defined(NCACHE_NXT) */

	len = Nc * sizeof(struct l_nch);
	if ((Ncache = (struct l_nch *)malloc((MALLOC_S)len)) == NULL) {

no_local_space:

	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: no space for %d local name cache\n", Pn, len);
	    exit(1);
	}
/*
 * Build a local copy of the kernel name cache.
 */

# if	defined(NCACHE_NXT)
	for (i = n = 0, kc = &nc, lc = Ncache; i < Nc && kp; i++)
# else	/* !defined(NCACHE_NXT) */
	for (i = n = 0, kc = kca, lc = Ncache; i < Nc; i++, kc++)
# endif	/* defined(NCACHE_NXT) */

	{

# if	defined(NCACHE_NXT)
		if (kread(kp, (char *)kc, sizeof(nc)))
			break;
		kp = (KA_T)kc->NCACHE_NXT;
# endif	/* defined(NCACHE_NXT) */

		if (!kc->NCACHE_NODEADDR)
			continue;
		if ((len = kc->NCACHE_NMLEN) < 1 || len > NCHNAMLEN)
			continue;
		if (len < 3 && kc->NCACHE_NM[0] == '.') {
			if (len == 1 || (len == 2 && kc->NCACHE_NM[1] == '.'))
				continue;
		}
		lc->na = (unsigned long)kc->NCACHE_NODEADDR;
		lc->id = kc->NCACHE_NODEID;

#  if	defined(NCACHE_PARADDR)
		lc->pa = (unsigned long)kc->NCACHE_PARADDR;
#  endif	/* defined(NCACHE_PARADDR) */

#  if	defined(NCACHE_PARID)
		lc->did = kc->NCACHE_PARID;
#  endif	/* defined(NCACHE_PARID) */

		(void) strncpy(lc->nm, kc->NCACHE_NM, len);
		lc->nm[len] = '\0';
		lc->nl = strlen(lc->nm);
		n++;
		lc++;
	}
/*
 * Reduce the local name cache memory allocation, if possible.
 * Free the space allocated to the kernel name cache copy.
 */

# if	!defined(NCACHE_NXT)
	(void) free((FREE_P *)kca);
# endif	/* !defined(NCACHE_NXT) */

	if (n < 1) {
		Nc = n;
		(void) free((FREE_P *)Ncache);
		Ncache = NULL;
		goto no_cache;
	}
	if (n < Nc) {
		Nc = n;
		len = Nc * sizeof(struct l_nch);
		if ((Ncache = (struct l_nch *)realloc(Ncache, len)) == NULL)
			goto no_local_space;
	}
/*
 * Build a hash table to locate Ncache entries.
 */
	for (Nch = 1; Nch < Nc; Nch <<= 1)
		;
	Nch <<= 1;
	Mch = Nch - 1;
	if ((Nchash = (struct l_nch **)calloc(Nch + Nc, sizeof(struct l_nch *)))
	== NULL) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: no space for %d name cache hash pointers\n",
		    Pn, Nch + Nc);
	    exit(1);
	}
	for (i = 0, lc = Ncache; i < Nc; i++, lc++) {
		for (hp = ncachehash(lc->id, lc->na), n = 1; *hp; hp++) {
			if ((*hp)->na == lc->na && (*hp)->id == lc->id
			&&  strcmp((*hp)->nm, lc->nm) == 0

# if	defined(NCACHE_PARADDR) && defined(NCACHE_PARID)
			&&  (*hp)->pa == lc->pa && (*hp)->did == lc->did
# endif	/* defined(NCACHE_PARADDR) && defined(NCACHE_PARID) */

			) {
				n = 0;
				break;
			}
		}
		if (n)
			*hp = lc;
	}

# if	defined(NCACHE_PARADDR) && defined(NCACHE_PARID)
/*
 * Make a final pass through the local cache and convert parent vnode
 * addresses to local name cache pointers.
 */
	for (i = 0, lc = Ncache; i < Nc; i++, lc++) {
		if (!lc->pa)
			continue;
		lc->pa = (unsigned long)ncache_addr(lc->did, lc->pa);
	}
# endif	/* defined(NCACHE_PARADDR) && defined(NCACHE_PARID) */
}


/*
 * ncache_lookup() - look up a node's name in the kernel's name cache
 */

char *
ncache_lookup(buf, blen)
	char *buf;			/* receiving name buffer */
	int blen;			/* receiving buffer length */
{
	char *cp;
	struct l_nch *lc;
	int nl, rlen;
/*
 * Look up the name cache entry for the node address.
 */
	if (Nc == 0 || (lc = ncache_addr(Lf->id, Lf->na)) == NULL)
		return(NULL);
	if ((nl = lc->nl) > (blen - 1))
		return(NULL);
	cp = buf + blen - nl - 1;
	rlen = blen - nl - 1;
	(void) strcpy(cp, lc->nm);

# if	defined(NCACHE_PARADDR) && defined(NCACHE_PARID)
/*
 * Look up the name cache entries that are parents of the node address.
 * Quit when:
 *
 *	there's no parent;
 *	the name length is too large to fit in the receiving buffer.
 */
	for (;;) {
		if (!(lc = (struct l_nch *)lc->pa)
		||  ((nl = lc->nl) + 1) > rlen)
			break;
		*(cp - 1) = '/';
		cp--;
		rlen--;
		(void) strncpy((cp - nl), lc->nm, nl);
		cp -= nl;
		rlen -= nl;
	}
# endif	/* defined(NCACHE_PARADDR) && defined(NCACHE_PARID) */
	return(cp);
}
#endif	/* defined(HASNCACHE) */
