/*
 * dproc.c - Linux process access functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dproc.c,v 1.5 95/02/23 09:43:11 abe Exp $";
#endif

#include "lsof.h"


_PROTOTYPE(static void get_kernel_access,(void));
_PROTOTYPE(static int nlist,(char *fnm, struct nlist *nl));
_PROTOTYPE(static void process_mapmem,(struct inode *x, struct vm_area_struct *m));


/*
 * Local static values
 */

static caddr_t *Nc = NULL;		/* node cache */
static MALLOC_S Nn = 0;			/* number of Nc[] entries allocated */


/*
 * gather_proc_info() -- gather process information
 */

void
gather_proc_info()
{
	struct file *f;
	struct file fs;
	struct task_struct **tt, *t, ts;
	int i, nt;
	MALLOC_S len;
	short pss, sf;
	uid_t uid;
/*
 * Allocate space for task structure pointers and read them.
 */
	len = (MALLOC_S)(NR_TASKS * sizeof(struct task_struct *));
	if ((tt = (struct task_struct **)malloc(len)) == NULL) {
		(void) fprintf(stderr,
			"%s: no space for %d task structure pointers\n",
			Pn, NR_TASKS);
		exit(1);
	}
	if (kread((KA_T)Nl[X_TASK].n_value, (char *)tt, (int)len)) {
		(void) fprintf(stderr,
			"%s: can't read task structure pointers from %#lx\n",
			Pn, Nl[X_TASK].n_value);
		exit(1);
	}
/*
 * Examine task structures and their associated information.
 */
	for (nt = 0, t = &ts; nt < NR_TASKS; nt++) {
		if (tt[nt] == (struct task_struct *)NULL
		||   kread((KA_T)tt[nt], (char *)t, sizeof(ts)))
			continue;
		if (t->state == TASK_ZOMBIE)
			continue;
		if (is_proc_excl((pid_t)t->pid, (pid_t)t->pgrp,
		    (UID_ARG)t->uid, &pss, &sf))
			continue;
	/*
	 * Allocate a local process structure.
	 */
		alloc_lproc(t->pid, t->pgrp, (UID_ARG)t->uid, t->comm,
			(int)pss, (int)sf);
		Plf = NULL;
	/*
	 * Save current working directory information.
	 */
		if (t->PWD) {
			alloc_lfile(CWD, -1);
			process_inode(t->PWD, (struct file_operations *)NULL);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Save root directory information.
	 */
		if (t->ROOT) {
			alloc_lfile(RTD, -1);
			process_inode(t->ROOT, (struct file_operations *)NULL);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Save information on the executable file and mapped memory.
	 */

#if	LINUXV<1147
		if (t->executable || t->mmap)
			process_mapmem(t->executable, t->mmap);
#else	/* LINUXV>=1147 */
		if (t->mm->mmap)
			process_mapmem(NULL, t->mm->mmap);
#endif	/* LINUXV<1147 */

	/*
	 * Save information on file descriptors.
	 */

		for (f = &fs, i = 0; i < NR_OPEN; i++) {

#if	LINUXV<1147
			if (t->filp[i] == (struct file *)NULL
			||  kread((KA_T)t->filp[i], (char *)f, sizeof(fs)))
#else	/* LINUXV>=1147 */
			if (t->files->fd[i] == (struct file *)NULL
			||  kread((KA_T)t->files->fd[i], (char *)f, sizeof(fs)))
#endif	/* LINUXV<1147 */

				continue;
			alloc_lfile(NULL, i);
			process_file(f);
			if (Lf->sf)
				link_lfile();
		}
	/*
	 * Examine results.
	 */
		if (examine_lproc())
			return;
	}
}


/*
 * get_kernel_access() - get access to kernel memory
 */

static void
get_kernel_access()
{
	if ((Core != NULL && !is_readable(Core))
	||  (Nmlst != NULL && !is_readable(Nmlst)))
		exit(1);
/*
 * Open kernel memory access.
 */
	if ((Kmem = open(Core ? Core : KMEM, O_RDONLY, 0)) < 0) {
		(void) fprintf(stderr, "%s: can't open %s: %s\n", Pn,
			Core ? Core : KMEM, strerror(errno));
		exit(1);
	}
/*
 * Access kernel symbols.
 */
	if (nlist(Nmlst ? Nmlst : N_UNIX, Nl) < 0) {
		(void) fprintf(stderr,
			"%s: can't read kernel name list from %s\n",
			Pn, Nmlst ? Nmlst : N_UNIX);
		exit(1);
	}
/*
 * Make sure important kernel symbols are defined.
 */
	if (Nl[X_TASK].n_value == (unsigned long int)NULL) {
		(void) fprintf(stderr,
			"%s: no _task kernel definition\n", Pn);
		exit(1);
	}
}


/*
 * initialize() - perform all initialization
 */

void
initialize()
{
	size_t sz;
	unsigned long hi, st;

	get_kernel_access();
	iuidcache(NR_TASKS);
}


/*
 * kread() - read from kernel memory
 */

int
kread(addr, buf, len)
	KA_T addr;			/* kernel memory address */
	char *buf;			/* buffer to receive data */
	READLEN_T len;			/* length to read */
{
	int br;

	if (lseek(Kmem, (off_t)addr, SEEK_SET) == (off_t)-1)
		return(-1);
	br = read(Kmem, buf, len);
	return((br == (int)len) ? 0 : 1);
}


/*
 * nlist() - get name list
 */

static int
nlist(fnm, nl)
	char *fnm;			/* name of file to search */
	struct nlist *nl;		/* names to locate */
{
	char buf[128];
	char *cp;
	FILE *fs;
	short nf, nn;
	struct nlist *np;

	if ((fs = fopen(fnm, "r")) == NULL)
		return(-1);
	for (nf = nn = 0, np = nl; np->n_name; nn++, np++)
		;
	while (nf < nn && fgets(buf, sizeof(buf), fs) != NULL) {
		if ((cp = strchr(buf, ' ')) == NULL)
			break;
		for (np = nl; np->n_name; np++) {
			if (np->n_value || strcmp(np->n_name, cp+3) != 0)
				continue;
			np->n_value = strtol(buf, &cp, 16);
			nf++;
			break;
		}
	}
	(void) fclose(fs);
	return(1);
}


/*
 * process_mapmem() - process mapped memory
 */

static void
process_mapmem(x, m)
	struct inode *x;
	struct vm_area_struct *m;
{
	int f;
	int i, j, lm;
	MALLOC_S ma;
	struct inode *ia;
	char *ty;
	struct vm_area_struct v;

	for (f = 1, i = lm = 0; m; lm++) {

	/*
	 * Avoid infinite loop.
	 */
		if (lm > 1000) {
			if (!Fwarn)
				(void) fprintf(stderr,
					"%s: too many memory maps, PID %d\n",
					Pn, Lp->pid);
				return;
		}
		if (f) {
		
		/*
		 * The first time through the loop, use the executable inode
		 * address.
		 */
			ty = "txt";
			f = 0;
			if ((ia = x) == (struct inode *)NULL)
				continue;
		} else {

		/*
		 * After the first time through the loop, use the inode
		 * addresses in the memory map.
		 */
			if (kread((KA_T)m, (char *)&v, sizeof(v)))
				break;
			m = v.vm_next;
			if ((ia = v.vm_inode) == (struct inode *)NULL)
				continue;
			ty = "mem";
		}
	/*
	 * Skip duplicate inode addresses.
	 */
		for (j = 0; j < i; j++) {
			if (Nc[j] == (caddr_t)ia)
				break;
		}
		if (j < i)
			continue;
	/*
	 * Cache the inode address for duplicate checking.
	 */
		if (i >= Nn) {
			Nn += 10;
			ma = (MALLOC_S)(Nn * sizeof(caddr_t));
			if (Nc)
				Nc = (caddr_t *)realloc(Nc, ma);
			else
				Nc = (caddr_t *)malloc(ma);
			if (!Nc) {
			    (void) fprintf(stderr,
				"%s: no space for memmap pointers, PID %d\n",
				Pn, Lp->pid);
			    exit(1);
			}
		}
		Nc[i++] = (caddr_t)ia;
	/* * Save the inode information.
	 */
		alloc_lfile(ty, -1);
		process_inode(ia, (struct file_operations *)NULL);
		if (Lf->sf)
			link_lfile();
	}
}

